classdef StabilityOptimizer < handle
    properties
		% Mesh data
        F
        V_outer
        V_inner
        V_skeleton
		
		% Bounds for vertex positions
        V_upper
        V_lower
        
        offset_directions
        
		% Full (truncated) basis, and reduced basis
        harmonics_basis
        reduced_basis
        
        use_subspace_projection
		
		% Current parameter bounds
        lb
        ub
        
		% Number of objective function evaluations
        obj_fun_evaluations = 0
		% Use this to save the mass moment gradient in the
		% objective function and read it in the out function
        last_gradient
		% Apply diagonal dominance out criterion in this iteration
        use_diag_dom_criterion
        % Diagonal dominance threshold
        global_out_threshold = 1
    end
    
    methods
        function obj = StabilityOptimizer()
        end
        
        function setModel(obj, mesh_path, skeleton_path, num_harmonics, min_thickness, model_height)
            % Load outer surface and skeleton from OBJ
			% readOBJ_mex is part of the libigl library
            [obj.V_outer, obj.F] = readOBJ_mex(mesh_path);
            obj.V_skeleton = readOBJ_mex(skeleton_path);
            
            % Scale along Y to specified height
            if nargin < 6
                initial_scaling = 1;
            else
                initial_scaling = model_height / range(obj.V_outer(:,2));
            end
            obj.V_outer = obj.V_outer * initial_scaling;
            obj.V_skeleton = obj.V_skeleton * initial_scaling;
            
            % Compute offset directions and maximum magnitudes
            outer_to_skeleton = obj.V_skeleton - obj.V_outer;
            obj.offset_directions = normr(outer_to_skeleton);
            max_offset_magnitude = sqrt(sum(outer_to_skeleton.^2,2));
            
            % Set lower and upper bounds for vertex positions
            obj.V_lower = obj.V_outer + obj.offset_directions * min_thickness;
            obj.V_upper = obj.V_outer + bsxfun(@times, obj.offset_directions, max_offset_magnitude);
            % Set initial inner surface as the mean between bounds
            obj.V_inner = 0.5 * (obj.V_lower + obj.V_upper);
            
            % Compute truncated harmonics basis
            % cotmatrix.m by Alec Jacobson (part of the GPToolbox library)
            L = cotmatrix(obj.V_outer, obj.F);        

            % compute k eigen-functions
            opts.v0 = ones(size(L, 1), 1);
            opts.issym = 1;
            [obj.harmonics_basis,~] = eigs(L, num_harmonics, 'sm', opts);
        end
        
        function performOptimization(obj, use_subspace_projection)
            obj.use_subspace_projection = use_subspace_projection;
            num_harmonics = size(obj.harmonics_basis, 2);
            
            % Initialize iteration data
            old_objective_value = Inf;
            do_full_iteration = false;
            num_reparametrizations = 0;
            initial_alpha = zeros(num_harmonics, 1);
            
            % Start timer
            tic;
            
            while true
                % Compute lower and upper bounds of full optimization
                % variables
                obj.ub = +sqrt(sum((obj.V_upper - obj.V_inner).^2,2));
                obj.lb = -sqrt(sum((obj.V_lower - obj.V_inner).^2,2));

                if use_subspace_projection
                    % Compute Jacobi matrix of relevant mass moments wrt to
                    % full optimization variables
                    num_reparametrizations = num_reparametrizations + 1;
                    gradV = obj.computeVertexGradient(initial_alpha, 0);

                    [~, grad_mom] = obj.calculateMoments();
                    delmom_delalpha = grad_mom * gradV;
                    relevant_grad = delmom_delalpha(2:4,:);
                    
                    % Find optimal parameter subspace
                    R = StabilityOptimizer.optimizeParametrization(relevant_grad);
                    R = normc(R);
                    obj.reduced_basis = obj.harmonics_basis*R;
                    basis = obj.reduced_basis;
                else
                    % Optimize in full space
                    basis = obj.harmonics_basis;
                end
                
                obj_fun = @(x) obj.evaluateObjective(x);
                initial_beta = zeros(size(basis,2), 1);
                
                if obj.use_subspace_projection
                    if do_full_iteration
                        % If we are almost done, do a full iteration
                        % without diagonal dominance out criterion
                        maxiter = 100;
                        outfun_ptr = [];
                        obj.use_diag_dom_criterion = 0;
                    else
                        maxiter = 20;
                        outfun_ptr = @obj.outFunction;
                        obj.use_diag_dom_criterion = 1;
                    end
                else
                    maxiter = 1000;
                    outfun_ptr = @obj.outFunction;
                    obj.use_diag_dom_criterion = 0;
                end
                
                if obj.use_subspace_projection
                    % Unconstrained optimization due to constraint mapping
                    options = optimoptions('fminunc', ...
                        'Algorithm', 'quasi-newton', ...
                        'HessUpdate','bfgs', ... %   bfgs', 'dfp', 'steepdesc'  % 'PlotFcns', {@optimplotx,@optimplotfval,@optimplotfirstorderopt}, ...                
                        'GradObj', 'on', ... %off on
                        'Display','iter-detailed', ...
                        'MaxIter', maxiter, ...
                        'TolFun', 10e-9, ...
                        'OutputFcn', outfun_ptr);

                    [final_beta, obj_val, ~] = fminunc(obj_fun, initial_beta, options);
                    obj.V_inner = obj.evaluateVertexPositions(final_beta, 1);
                else
                    % Full optimization without reduced basis and without
                    % constraint mapping
                    A_lin = [basis; -basis];
                    b_lin = [obj.ub; -obj.lb];

                    options = optimoptions('fmincon', ...
                        'Algorithm', 'active-set', ...         
                        'GradObj', 'on', ... %off on
                        'Display','iter-detailed', ...
                        'MaxIter', maxiter, ...
                        'TolFun', 10e-9, ...
                        'OutputFcn', outfun_ptr);

                    [final_beta, obj_val, ~] = fmincon(obj_fun, initial_beta, A_lin, b_lin, [], [], [], [], [], options);
                    obj.V_inner = obj.evaluateVertexPositions(final_beta, 0);
                end
                
                if ~obj.use_subspace_projection
                    break;
                end

                if (old_objective_value - obj_val)/obj_val < 0.01
                    % If change in objective value was less than 1%, do an
                    % iteration without using the diagonal dominance out
                    % criterion
                    if do_full_iteration
                        % If we already did that in the last iteration,
                        % we're done
                        break;
                    else
                        do_full_iteration = true;
                    end
                else
                    do_full_iteration = false;
                end

                old_objective_value = obj_val;
            end
            
            % Stop timer
            time_elapsed = toc;
            
            % Plot the result
            obj.plotSurfaces();
            fprintf('Total # function evaluations: %d\n', obj.obj_fun_evaluations);
            fprintf(' Total time spent optimizing: %.2f seconds\n', time_elapsed);
            fprintf('             Objective value: %.4e\n', obj_val);
        end
        
        function [objective, grad_objective] = evaluateObjective(obj, params)
            % Evaluate objective function
            obj.obj_fun_evaluations = obj.obj_fun_evaluations + 1;

            % Compute vertex positions, moments, and their gradients
            V_inner_cur = obj.evaluateVertexPositions(params, obj.use_subspace_projection);
            delV_delalpha = obj.computeVertexGradient(params, obj.use_subspace_projection);
            [moments, delmoments_delV] = obj.calculateMoments(V_inner_cur);
            delmoments_delalpha = delmoments_delV * delV_delalpha;

            c = moments(2:4);
            grad_c = delmoments_delalpha(2:4, :);
            obj.last_gradient = grad_c;

            % Objective is weighted sum of relevant mass moment derivatives
            objective = 1e4*(c(1)^2 + c(3)^2) + c(2);
            grad_objective = 2*1e4*(c(1)*grad_c(1,:) + c(3)*grad_c(3,:)) + grad_c(2,:);
        end
        
        function ev = evaluateVertexPositions(obj, params, reduced)
            % Compute vertex positions based on current parameter values
            if reduced
                basis = obj.reduced_basis;
            else
                basis = obj.harmonics_basis;
            end
            
            if obj.use_subspace_projection
                % Account for constraint mapping
                o = tan(pi*(obj.ub+obj.lb) ./ (2*(obj.lb-obj.ub)));
                delta = basis * params;
                delta2 = atan(delta + o) .* (obj.ub-obj.lb)/pi + (obj.ub+obj.lb)*0.5;
                ev = obj.V_inner + bsxfun(@times, obj.offset_directions, delta2);
            else
                ev = obj.V_inner + bsxfun(@times, obj.offset_directions, basis * params);
            end
        end
        
        function stop = outFunction(obj, ~, ~, state)
            stop = false;
            if strcmp(state, 'iter')
                if obj.use_diag_dom_criterion
                    % Check if Jacobi matrix of user goals (mass moments)
                    % wrt reduced parameters is still diagonally dominant.
                    % Otherwise stop optimizing and reparametrize.
                    absDiag  = abs(diag(obj.last_gradient));
                    absElem = sum(abs(obj.last_gradient), 2) - absDiag;
                    d = sum(absElem./absDiag)/ length(absDiag);

                    stop = d > obj.global_out_threshold;
                end
            end
        end
        
        function gradV = computeVertexGradient(obj, params, reduced)
            % Computes derivatives of vertex positions wrt the harmonics
            % parameters
            if reduced
                basis = obj.reduced_basis;
            else
                basis = obj.harmonics_basis;
            end
            
            np = size(basis, 2);
            nv = size(basis, 1);
            if obj.use_subspace_projection
                % Include influence of constraint mapping
                oi = tan(pi*(obj.ub+obj.lb)./(2*(obj.lb-obj.ub)));
                mult = (obj.ub-obj.lb)/pi;
                temp = (1 ./ (1 + (basis * params + oi).^2)) .* mult;
                inner = bsxfun(@times, basis, temp);
                inner_gradV = bsxfun(@times, repmat(inner, 3, 1), obj.offset_directions(:));
            else
                inner_gradV = bsxfun(@times, repmat(basis, 3, 1), obj.offset_directions(:));
            end

            gradV = [zeros(nv,np); inner_gradV(1:nv,:); ...
                zeros(nv,np); inner_gradV((nv+1):(2*nv),:); ...
                zeros(nv,np); inner_gradV((2*nv+1):(3*nv),:)];
        end
        
        function [mom, grad_mom] = calculateMoments(obj, V)
            % Compute mass moments and their gradients
            if nargin < 2
                V = obj.V_inner;
            end
            
            nv = size(obj.V_outer, 1);

            V_all = [obj.V_outer; V];
            F_all = [obj.F; obj.F(:,[1 3 2]) + nv];

            t1 = V_all(F_all(:,1),:);
            t2 = V_all(F_all(:,2),:);
            t3 = V_all(F_all(:,3),:);

            [mom,grad_mom] = comp_moments_grad_3d(t1, t2, t3, F_all, zeros(size(V_all)));
            grad_mom = [grad_mom(:,1:3:end) grad_mom(:,2:3:end) grad_mom(:,3:3:end)];
        end
        
        function plotSurfaces(obj)
            figure;
            hold on;
            patch('Vertices',obj.V_inner,'Faces',obj.F,'FaceColor',[0.4 0.4 0.8],'EdgeColor',[0.6 0.6 0.9]);
            patch('Vertices',obj.V_outer,'Faces',obj.F,'FaceColor',[0.8 0.4 0.4],'FaceAlpha',0.4,'EdgeColor','none');
            axis equal tight vis3d;
        end
    end
    
    methods(Static)
        function ret = optimizeParametrization(grad_response)
            % Compute min_B ||B|| s.t. grad_response * B = I
            np = size(grad_response, 2);
            nf = size(grad_response, 1);

            C_vec = kron(eye(nf), grad_response);
            d = eye(nf);
            d = d(:);
            A = eye(nf*np);
            b = zeros(nf*np, 1);

            KKT = [A'*A C_vec'; C_vec zeros(nf*nf)];
            rhs = [A'*b; d];

            ret = KKT\rhs;
            ret = reshape(ret(1:(np*nf)), np, []);
        end
    end
end